#include "veh_measure.h"

namespace Localization {
namespace Ego {

double VehMeasure::latestUpdateTime_ = -1.0;

VehMeasure::VehMeasure(const MeasurementType &mType, const VehInfo &vehInfo)
:BaseMeasure(mType)
{
  isVehBlock_ = false;
  stamp_ = vehInfo.stamp_;
  if (stamp_ - latestUpdateTime_ < 1e-5) {
    isVehBlock_ = true;
    std::cout<<std::setprecision(15)<<"VehMeasure():this frame is repeated,stamp"<<stamp_
        <<",latestUpdateTime_"<<latestUpdateTime_<<std::endl;
  }
  latestUpdateTime_ = stamp_;

  seq_ = vehInfo.seq_;
  type_ = mType;
  isKFUsed_ = false;

  yawRate_    = vehInfo.yawRate_;
  wheelSpeed_ = vehInfo.velocity_;
  vm_ = Eigen::Vector3d(wheelSpeed_, 0, 0);
  wm_ = Eigen::Vector3d(0, 0, yawRate_);
  isStatic_ = IsStatic();
  vehVelo_ = wheelSpeed_;

  mRm_ = Eigen::MatrixXd::Zero(6, 6);
  mRm_(0, 0) = wheelVelocityVarX_;
  mRm_(1, 1) = wheelVelocityVarY_;
  mRm_(2, 2) = wheelVelocityVarZ_;
  mRm_.block<3, 3>(3, 3) = wheelAngleRateVar_ * Eigen::Matrix3d::Identity();
  CheckValid();
}

void VehMeasure::CheckValid()
{
  isVehValid_ = true;
  lastVehStamp_ = stamp_;
  isVehBlock_ = false;
}

bool VehMeasure::IsStatic()
{
  if (wheelSpeed_ < 1e-3) {
    return true;
  } else {
    return false;
  }
}

bool VehMeasure::Update(const State &state, State &nextState)
{
  nextState = state;
  if (isVehValid_ == false) {
    std::cout<<"veh-Update():isVehValid_ false,can't update,use next pridict"<<std::endl;
    return false;
  }
  nextState.time = stamp_;

  if (isStatic_) {
    vm_ = Eigen::Vector3d(0.0, 0.0, 0.0);
    wm_ = Eigen::Vector3d(0.0, 0.0, 0.0);
    mRm_ = 0.001 * mRm_;
  }

  Eigen::VectorXd innov(6);
  Eigen::MatrixXd mH = Eigen::MatrixXd::Zero(6, StateIndex::STATE_TOTAL);

  ObserveVeh(state,nextState,innov,mH);

  (void)UpdateStateWheel(state, nextState, innov, mH);

  return true;
}

bool VehMeasure::UpdateStateWheel(const State &state, State &nextState, const Eigen::VectorXd &innov,
     const Eigen::MatrixXd &mH)
{
  Eigen::MatrixXd mS = mH * state.mP * mH.transpose() + mRm_;
  Eigen::MatrixXd mK = state.mP * mH.transpose() * mS.inverse();
  Eigen::VectorXd dx = mK * innov;

  UpdateState(state, nextState, dx, mH, mK);
  return true;
}

Eigen::VectorXd VehMeasure::MeasurementEquation(const State &state) const
{
  Eigen::VectorXd measurement(6);
  Eigen::VectorXd tempMeasure(3);
  measurement = tempMeasure;
  measurement = state.ws * (state.q.inverse() * state.v);

  return measurement;
}

void VehMeasure::ObserveVeh(const State &state, State &nextState, Eigen::VectorXd &innov,
                                            Eigen::MatrixXd &mH)
{
  Eigen::VectorXd tempInnov(3);
  innov = tempInnov;
  Eigen::MatrixXd tempH = Eigen::MatrixXd::Zero(3, StateIndex::STATE_TOTAL);
  mH = tempH;

  innov = vm_ - state.ws * (state.q.inverse() * state.v);
  mH.block<3, 3>(0, StateIndex::R) = SkewSymmetric(state.ws * (state.q.inverse() * state.v));
  mH.block<3, 3>(0, StateIndex::V) = state.ws * (state.q.toRotationMatrix().transpose());

  Eigen::MatrixXd tmpR = mRm_.block<3, 3>(0, 0);
  mRm_ = tmpR;
}

} // namespace Ego
} // namespace Localization
